
/*
 * Name: main.c
 * Project: arm-sensor
 * Author: kawamoto@devdrv.co.jp
 * Creation Date: 2009-12-01
 */

#include "type.h"
#include "debug.h"

#include "hal.h"
#include "console.h"
#include "usbapi.h"
#include "lpc23xx.h"

#define INTR_IN_EP 0x81
#define MAX_PACKET_SIZE 8

#define BAUD_RATE 115200

#define LE_WORD(x) ((x)&0xFF),((x)>>8)

static U8 abClassReqData[4];
static U8 abReport[MAX_PACKET_SIZE];

// see the joystick example from the usb.org HID Descriptor Tool
static const U8 abReportDesc[] =
{
	0x06, 0x00, 0xFF,		// USAGE_PAGE (Vendor Defined Page 1)
	0x09, 0x01,				// USAGE (Vendor Usage 1)
	0xA1, 0x01,				// COLLECTION (Application)
	0x15, 0x00,				//   LOGICAL_MINIMUM (0)
	0x26, 0xFF, 0x00,		//   LOGICAL_MAXIMUM (255)
	0x75, 0x08,				//   REPORT_SIZE (8)
	0x95, 0x08,				//   REPORT_COUNT (8)
	0x09, 0x01,				//   USAGE (Vendor Usage 1)
	0x81, 0x02,				//   INPUT (Data,Var,Abs)
	0xC0					// END_COLLECTION
};

static const U8 abDescriptors[] =
{
/* Device descriptor */
	0x12,              		
	DESC_DEVICE,       		
	LE_WORD(0x0200),		// bcdUSB
	0x00,              		// bDeviceClass
	0x00,              		// bDeviceSubClass
	0x00,              		// bDeviceProtocol
	MAX_PACKET_SIZE0,  		// bMaxPacketSize
	LE_WORD(0x1384),		// idVendor
	LE_WORD(0x0200),		// idProduct
	LE_WORD(0x0100),		// bcdDevice
	0x01,              		// iManufacturer
	0x02,              		// iProduct
	0x03,              		// iSerialNumber
	0x01,              		// bNumConfigurations

// configuration
	0x09,
	DESC_CONFIGURATION,
	LE_WORD(0x22),  		// wTotalLength
	0x01,  					// bNumInterfaces
	0x01,  					// bConfigurationValue
	0x00,  					// iConfiguration
	0x80,  					// bmAttributes
	50, 					// bMaxPower(100mA)

// interface
	0x09,   				
	DESC_INTERFACE, 
	0x00,  		 			// bInterfaceNumber
	0x00,   				// bAlternateSetting
	0x01,   				// bNumEndPoints
	0x03,   				// bInterfaceClass = HID
	0x00,   				// bInterfaceSubClass
	0x00,   				// bInterfaceProtocol
	0x00,   				// iInterface

// HID descriptor
	0x09, 
	DESC_HID_HID, 			// bDescriptorType = HID
	LE_WORD(0x0111),		// bcdHID
	0x00,   				// bCountryCode
	0x01,   				// bNumDescriptors = report
	DESC_HID_REPORT,   		// bDescriptorType
	LE_WORD(sizeof(abReportDesc)),

// EP descriptor
	0x07,   		
	DESC_ENDPOINT,   		
	INTR_IN_EP,				// bEndpointAddress
	0x03,   				// bmAttributes = INT
	LE_WORD(MAX_PACKET_SIZE),// wMaxPacketSize
	10,						// bInterval

// string descriptors
	0x04,
	DESC_STRING,
	LE_WORD(0x0409),

	// manufacturer string
	0x0E,
	DESC_STRING,
	'D', 0, 'e', 0, 'v', 0, 'D', 0, 'r', 0, 'v', 0,

	// product string
	0x0E,
	DESC_STRING,
	'S', 0, 'e', 0, 'n', 0, 's', 0, 'o', 0, 'r', 0, 

	// serial number string
	0x0E,
	DESC_STRING,
	'S', 0, 'a', 0, 'm', 0, 'p', 0, 'l', 0, 'e', 0,
	
	// terminator
	0
};


/*************************************************************************
	HandleClassRequest
	==================
		HID class request handler
		
**************************************************************************/
static BOOL HandleClassRequest(TSetupPacket *pSetup, int *piLen, U8 **ppbData)
{
	U8	*pbData = *ppbData;

	switch(pSetup->bRequest)
	{
		case HID_GET_REPORT:
			DBG("GET REPORT, val=%X, idx=%X\n", pSetup->wValue, pSetup->wIndex);
			pbData = abReport;
			*piLen = sizeof(abReport);
			break;
		default:
			DBG("Unhandled class %X\n", pSetup->bRequest);
			return(FALSE);
	}
	return(TRUE);
}


/*************************************************************************
	HIDHandleStdReq
	===============
		Standard request handler for HID devices.
		
	This function tries to service any HID specific requests.
		
**************************************************************************/
static BOOL HIDHandleStdReq(TSetupPacket *pSetup, int *piLen, U8 **ppbData)
{
	U8 bType, bIndex;

	if((pSetup->bmRequestType == 0x81) &&			// standard IN request for interface
		(pSetup->bRequest == REQ_GET_DESCRIPTOR))	// get descriptor
	{
		bType = GET_DESC_TYPE(pSetup->wValue);
		bIndex = GET_DESC_INDEX(pSetup->wValue);
		switch(bType)
		{
			case DESC_HID_REPORT:
				// report
				*ppbData = abReportDesc;
				*piLen = sizeof(abReportDesc);
				break;
			case DESC_HID_HID:
			case DESC_HID_PHYSICAL:
			default:
				// search descriptor space
				return(USBGetDescriptor(pSetup->wValue, pSetup->wIndex, piLen, ppbData));
		}
		
		return(TRUE);
	}
	return(FALSE);
}


static int chg_flg = 0;
static unsigned char ax = 0, ay = 0, az = 0;
static unsigned char lx = 0;
static unsigned char sw = 0;

#define FLG_ACCELERATION (1 << 0)
#define FLG_ILLUMINANCE  (1 << 1)
#define FLG_SWITCH       (1 << 2)

#define HDR_NULL         ((0 << 3) | 0)	/* ID = 0, LEN = 0 */
#define HDR_ACCELERATION ((1 << 3) | 3)	/* ID = 1, LEN = 3 */
#define HDR_ILLUMINANCE  ((2 << 3) | 1)	/* ID = 2, LEN = 1 */
#define HDR_SWITCH       ((3 << 3) | 1)	/* ID = 3, LEN = 1 */

static void IntrIn(U8 bEP, U8 bEPStatus)
{
	unsigned char new_ax, new_ay, new_az;
	unsigned char new_lx;
	unsigned char new_sw;

	if(chg_flg == 0)
	{
		AD0CR = (AD0CR & ~0xFF) | (1 << 0);			// AD0_0(KXM52-1050 OUTX)
		AD0CR = (AD0CR & ~(7 << 24)) | (1 << 24);	// START = 1
		while(!(AD0GDR & 0x80000000));
		AD0CR = (AD0CR & ~(7 << 24));				// START = 0
		new_ax = (AD0GDR & 0xFF00) >> 8;

		AD0CR = (AD0CR & ~0xFF) | (1 << 1);			// AD0_1(KXM52-1050 OUTY)
		AD0CR = (AD0CR & ~(7 << 24)) | (1 << 24);	// START = 1
		while(!(AD0GDR & 0x80000000));
		AD0CR = (AD0CR & ~(7 << 24));				// START = 0
		new_ay = (AD0GDR & 0xFF00) >> 8;

		AD0CR = (AD0CR & ~0xFF) | (1 << 2);			// AD0_2(KXM52-1050 OUTZ)
		AD0CR = (AD0CR & ~(7 << 24)) | (1 << 24);	// START = 1
		while(!(AD0GDR & 0x80000000));
		AD0CR = (AD0CR & ~(7 << 24));				// START = 0
		new_az = (AD0GDR & 0xFF00) >> 8;

		AD0CR = (AD0CR & ~0xFF) | (1 << 3);			// AD0_3(NJL7502L)
		AD0CR = (AD0CR & ~(7 << 24)) | (1 << 24);	// START = 1
		while(!(AD0GDR & 0x80000000));
		AD0CR = (AD0CR & ~(7 << 24));				// START = 0
		new_lx = (AD0GDR & 0xFF00) >> 8;

		new_sw = ((FIO0PIN & (1 << 27) ? 0:1) << 7)			// P0_27(SW8)
					| ((FIO0PIN & (1 << 17) ? 0:1) << 6)	// P0_17(SW7)
					| ((FIO0PIN & (1 << 9) ? 0:1) << 5)		// P0_9(SW6)
					| ((FIO0PIN & (1 << 8) ? 0:1) << 4)		// P0_8(SW5)
					| ((FIO0PIN & (1 << 7) ? 0:1) << 3)		// P0_7(SW4)
					| ((FIO0PIN & (1 << 6) ? 0:1) << 2)		// P0_6(SW3)
					| ((FIO0PIN & (1 << 5) ? 0:1) << 1)		// P0_5(SW2)
					| ((FIO0PIN & (1 << 4) ? 0:1) << 0);	// P0_4(SW1)

		if((ax != new_ax) || (ay != new_ay) || (az != new_az))
		{
			chg_flg = chg_flg | FLG_ACCELERATION;
			ax = new_ax; ay = new_ay; az = new_az;
		}

		if(lx != new_lx)
		{
			chg_flg = chg_flg | FLG_ILLUMINANCE;
			lx = new_lx;
		}

		if(sw != new_sw)
		{
			chg_flg = chg_flg | FLG_SWITCH;
			sw = new_sw;
		}
	}

	if(chg_flg & FLG_ACCELERATION)
	{
		abReport[0] = HDR_ACCELERATION;
		abReport[1] = ax;
		abReport[2] = ay;
		abReport[3] = az;
		
		chg_flg = chg_flg & ~FLG_ACCELERATION;
	}
	else if(chg_flg & FLG_ILLUMINANCE)
	{
		abReport[0] = HDR_ILLUMINANCE;
		abReport[1] = lx;
		
		chg_flg = chg_flg & ~FLG_ILLUMINANCE;
	}
	else if(chg_flg & FLG_SWITCH)
	{
		abReport[0] = HDR_SWITCH;
		abReport[1] = sw;
		
		chg_flg = chg_flg & ~FLG_SWITCH;
	}
	else
	{
		abReport[0] = HDR_NULL;
	}

	USBHwEPWrite(INTR_IN_EP, abReport, sizeof(abReport));
}


/*************************************************************************
	main
	====
**************************************************************************/
int main(void)
{
	// PLL and MAM
	HalSysInit();

	// init DBG
	ConsoleInit(24000000 / (16 * BAUD_RATE));
//	ConsoleInit(72000000 / (16 * BAUD_RATE));

	// init GPIO
	// 1111 1111 1111 0000 0000 1100 0000 0000
	// 5544 3322 1100 9988 7766 5544 3322 1100
	// XXXX XXXX XXXX 0000 0000 0000 XXXX XXXX
	PINSEL0 = PINSEL0 & 0xFFF000FF;					// P0_9 - P0_4
	
	// 3333 2222 2222 2222 2222 2222 1111 1111
	// 1100 9988 7766 5544 3322 1100 9988 7766
	// XXXX XXXX 00XX XXXX XXXX XXXX XXXX 00XX
	PINSEL1 = PINSEL1 & 0xFF3FFFF3;					// P0_27, P0_17

	// 1111 1111 1111 0000 0000 1100 0000 0000
	// 5544 3322 1100 9988 7766 5544 3322 1100
	// XXXX XXXX XXXX XXXX 0000 0000 0000 0000
	PINSEL4 = PINSEL4 & 0xFFFF0000;					// P2_7 - P2_0

	// init ADC
	PCONP    = PCONP | (1 << 12);
	PCLKSEL0 = (PCLKSEL0 & ~(3 << 24)) | (1 << 24);	// PCLK_ADC = 1(24MHz)
	AD0CR    = (AD0CR & ~(0xFF << 8)) | (6 << 8);	// CLKDIV = 6(4MHz)
	AD0CR    = AD0CR & ~(0x0F << 16);				// CLKS = 0, BURST = 0
	AD0CR    = AD0CR | (1 << 21);					// PDN = 1
	
	// 3333 2222 2222 2222 2222 2222 1111 1111
	// 1100 9988 7766 5544 3322 1100 9988 7766
	// XXXX XXXX XX01 0101 01XX XXXX XXXX XXXX
	PINSEL1  = (PINSEL1 & 0xFFC03FFF) | 0x00154000;	// AD0_3 - AD0_0

	DBG("Initialising USB stack\n");

	// initialise stack
	USBInit();
	USBHwNakIntEnable(1 << 3);

	// register device descriptors
	USBRegisterDescriptors(abDescriptors);

	// register HID standard request handler
	USBRegisterCustomReqHandler(HIDHandleStdReq);

	// register class request handler
	USBRegisterRequestHandler(REQTYPE_TYPE_CLASS, HandleClassRequest, abClassReqData);

	// register endpoint
	USBHwRegisterEPIntHandler(INTR_IN_EP, IntrIn);

	DBG("Starting USB communication\n");

	// connect to bus
	USBHwConnect(TRUE);

	// call USB interrupt handler continuously
	while(1)
	{
		USBHwISR();
	}

	return(0);
}

